import numpy as nmp 

class ConformationVector(object):
	""" A single column vector of certain length and binary in nature that will tell the configuration of a certain instance of
	the ConformationEntity Class at some instant of time """

	def __init__(self, arg_vector):
		""" I am using quantum mechanics lingo here for fun """
		self.ket = bytearray(arg_vector)
		self.order = len(arg_vector)


class ConformationEntity(object):
	""" A class depicting an entity whose configurations and their distribution will be measured to estimate the probabilities 
	for computing the topographical entropy. An object of this class could be a Dihedral angle, an amino acids backbone phi/psi 
	angles, a whole protein, etc. This object will have a conformation/configuration vector (binary in 0 or 1) at each frame that
	will tell its status at that frame. E.g. A dihedral can be described by a conformation vector ordered as |g+, g-, t> and at some 
	time 't' its values is |0, 0, 1>. This means that at that time, te dihedral was in 'trans' state. Same can be done for a residue
	in a protein with flags indicating which quadrant in the Ramachandran plot the residue's backbone was in. """

	def __init__(self, arg_order, arg_numFrames, **kwargs):
		self.timeSeries = -1 * nmp.ones((arg_order, arg_numFrames))   # a 2D array where each column represents a ConformationVector at a certain time/snapshot (initialized with absurd -1's).
		for k, v in kwargs.items():
			setattr(self, k, v)
			# print('{} -> {}'.format(k, v))

		







import numpy as nmp
from CodeEntropy.FunctionCollection import Utils
from CodeEntropy.Trajectory import TrajectoryConstants as TCON

def sort_modes(arg_modeSpectra):
	"""
	Sort the modes in ascending order using mode frequencies
	and return a new sorted spectrum
	"""
	retSpectra = []
	ascOrder = nmp.argsort([m for m in arg_modeSpectra])
	for im, midx in enumerate(ascOrder):
		arg_modeSpectra[midx].modeIdx = im + 1
		retSpectra.append(arg_modeSpectra[midx])

	return retSpectra
#END

class Mode(object):
	"""
	A class that contains information about a 
	fundamental mode of motion.
	"""

	def __init__(self, arg_modeIdx = 0, arg_modeEval = -1,\
	             arg_modeEvec = [], arg_modeFreq = -1,  **kwargs):
		self.modeIdx = arg_modeIdx
		self.modeEval = arg_modeEval
		self.modeEvec = arg_modeEvec
		self.modeFreq = arg_modeFreq

		for k, v in kwargs.items():
			setattr(self, k, v)

		self.modeAmpl = 0

	#END

	def __lt__(self, otherMode):
		return self.modeFreq < otherMode.modeFreq

	def __gt__(self, otherMode):
		return self.modeFreq > otherMode.modeFreq

	def __eq__(self, otherMode):
		return self.modeFreq == otherMode.modeFreq

	def get_mode_vecorder(self):
		return len(self.modeEvec)

	def get_line_nmd(self, arg_wfac):
		""" returns a string containing the mode info in the format
		suitable for NMWiz (NMD file)"""

		# mode - Normal mode array. Each normal mode array must be provided 
		# in one line as a list of decimal numbers. 
		# Mode array may be preceded by mode index and mode length (scaling factor) 
		# For FC: scaling factor is KT/sqrt(L_i*m_j) for mode `i` and bead `j` (mass weighting)
		# For TC: scaling factor is KT/sqrt(L_i*I_jj) for mode `i` and bead `j` (intertia weighting)
		# For NMA, QH, etc. this is different.
		#
		# For FC and TC, a common mass-independent mode amplitude factor is set.
		# therefore, when printing the eigenvectors out, they must be mass weighted (FC) or
		# inertia weighted for (TC).
		#
		# `arg_wfac` is an array of weighting factors
		#
		# Outpur format: "Mode %d %.2f x3 x3 ... x3"
		try:
			assert(int(self.get_mode_vecorder()/TCON.VECTORDIM) == len(arg_wfac))
		except:
			raise ValueError("Eigen vector and weight factor array are of different lengths ({} vs. {})"\
				.format(int(self.get_mode_vecorder()/TCON.VECTORDIM), len(arg_wfac)))

		nrow = int(self.get_mode_vecorder()/TCON.VECTORDIM)
		retStr = "Mode {:d} {:.2f} ".format(self.modeIdx, self.modeAmpl)
		evecs = nmp.reshape(a = self.modeEvec,\
			           newshape = (nrow, TCON.VECTORDIM))

		for uv, um in zip(evecs, arg_wfac):
			for uvm in nmp.divide(uv, nmp.sqrt(um)):
				retStr = "{} {:.3f}".format(retStr, uvm)

		return retStr
	#END


import numpy as nmp
from CodeEntropy.Trajectory import TrajectoryFrame as TF
from CodeEntropy.Trajectory import TrajectoryConstants as TCON
from CodeEntropy.FunctionCollection import Utils
from CodeEntropy.ClassCollection import BondStructs

#warining
class ComplexWarning(Warning):
	"""Waring for complex number result
	"""
	pass

class DataContainer(object):
	""" 
	This is the main data container for CodeEntropy Solute calculation
	This host framewise information of positions and forces on all the atoms, in lab and local frames. The vectors in the local frames are obtained after transforming the lab frame vectors  using orthonormal bases that also stored framewise for each atom. 

	The properties of its molecule (base molecule) is linked to a molecule class object that contains the info about its topology. 
	"""

	def __init__(self, u, start=None, end=None, step=1):
		"""
		Create a data container for CodeEntropy and load data from MDAnalysis.Universe
		
		Parameters
		----------
		u : MDAnalyse.Universe
			A Universe object will all topology, dihedrals,coordinates and force information Check ``Example/create_new_universe.py`` on how to create a universe from existing data
		start : int or None, Optional, default: None
			Frame id to start analysis. Default None will start from frame 0
		end : int or None, Optional, default: None
			Frame id to end analysis. Default None will end at last frame
		step : int, Optional, default: 1
			Steps between frame.
		"""
		self.universe = u
		self.numFrames = len(self.universe.trajectory)
		self.frameIndices = []     # a list of integer values
		self.trajSnapshots = []  # a list of instances of TrajectoryFrame class

		num_atom = self.universe.atoms.n_atoms

		self.isCAtomArray = nmp.zeros(num_atom)
		self.isOAtomArray = nmp.zeros(num_atom)
		self.isNAtomArray = nmp.zeros(num_atom)
		self.isCaAtomArray = nmp.zeros(num_atom)
		self.isBBAtomArray = nmp.zeros(num_atom)
		self.isHydrogenArray = nmp.zeros(num_atom)

		selection_pair = [
			(self.isCAtomArray, "name C"),
			(self.isOAtomArray, "name O"),
			(self.isNAtomArray, "name N"),
			(self.isCaAtomArray, "name CA"),
			(self.isBBAtomArray, "backbone"),
			(self.isHydrogenArray, "name H*")
		]
		for item in selection_pair:
			idx_list = u.atoms.select_atoms(item[1]).indices
			for id in idx_list:
				item[0][id] = 1

		self.dihedralArray = set()
		self.dihedralTable = dict()
		dihedList = u.dihedrals.indices
		for i in range(num_atom):
			self.dihedralTable[i] = set()
		
		for dihedrals in dihedList:
			newDih = BondStructs.Dihedral(dihedrals, self)
			self.add_dihedral(newDih)

		#reading trajectorys into memory because MDanalysis reads values on the fly which might slow down processing speed as these values are accessed multiple times
		if start == None:
			start = 0
		if end == None:
			end = len(u.trajectory)
		for i in range(int(start), int(end), int(step)):
			ts = u.trajectory[i]
			newFrame = TF.TrajectoryFrame(arg_frameIndex = ts.frame, \
													arg_vectorDim = TCON.VECTORDIM)
			newFrame.set_numAtoms(ts.n_atoms)
			newFrame.set_timeStep(ts.time)
			newFrame.initialize_matrices(arg_hasVelocities = ts.has_velocities, arg_hasForces = ts.has_forces)
			newFrame.value["coordinates"] = nmp.array(ts.positions)
			if ts.has_velocities:
				newFrame.value["velocities"] = nmp.array(ts.velocities)
			if ts.has_forces:
				newFrame.value["forces"] = nmp.array(ts.forces)
			self.trajSnapshots.append(newFrame)
			self.frameIndices.append(ts.frame)

		self.print_attributes()

		self._labCoords = nmp.ndarray( (len(self.trajSnapshots), self.universe.atoms.n_atoms, 3) )
		self._labForces = nmp.ndarray( (len(self.trajSnapshots), self.universe.atoms.n_atoms, 3) )
		
		self.translationAxesArray = nmp.ndarray ( (len(self.trajSnapshots), self.universe.atoms.n_atoms, 1+3, 3) ) # 4rth row is the origin
		self.rotationAxesArray = nmp.ndarray ( (len(self.trajSnapshots), self.universe.atoms.n_atoms, 1+3, 3) ) # 4rth row is the origin

		self.localCoords = nmp.ndarray( (len(self.trajSnapshots), self.universe.atoms.n_atoms, 3) )
		self.localForces = nmp.ndarray( (len(self.trajSnapshots), self.universe.atoms.n_atoms, 3) )
		self.localTorques = nmp.ndarray( (len(self.trajSnapshots), self.universe.atoms.n_atoms, 3) )
		# read the coords and forces from the trajectory
		# and store them in the mainContainer
		# it is a gromacs trajectory
		self.initialize_ndarrays()

		for i in range(len(self.trajSnapshots)):
			self._labCoords[i] = self.trajSnapshots[i].value['coordinates']
			self._labForces[i] = self.trajSnapshots[i].value['forces']

		
	def print_attributes(self):
		# print("{:<20s} : {}".format("Molecule name", self.molecule.name))
		print("{:<20s} : {}".format("Number of atoms", self.universe.atoms.n_atoms))
		print("{:<20s} : {}".format("Number of frames", len(self.trajSnapshots)))
		
		return
  
	def initialize_ndarrays(self):
		
		""" The number of frames for which the container will hold the data must be provided before hand"""
		assert(len(self.trajSnapshots) >= 0 and len(self.frameIndices) == len(self.trajSnapshots))

		self._labCoords = nmp.ndarray( (len(self.trajSnapshots), self.universe.atoms.n_atoms, 3) )
		self._labForces = nmp.ndarray( (len(self.trajSnapshots), self.universe.atoms.n_atoms, 3) )
		
		self.translationAxesArray = nmp.ndarray ( (len(self.trajSnapshots), self.universe.atoms.n_atoms, 1+3, 3) ) # 4rth row is the origin
		self.rotationAxesArray = nmp.ndarray ( (len(self.trajSnapshots), self.universe.atoms.n_atoms, 1+3, 3) ) # 4rth row is the origin

		self.localCoords = nmp.ndarray( (len(self.trajSnapshots), self.universe.atoms.n_atoms, 3) )
		self.localForces = nmp.ndarray( (len(self.trajSnapshots), self.universe.atoms.n_atoms, 3) )
		self.localTorques = nmp.ndarray( (len(self.trajSnapshots), self.universe.atoms.n_atoms, 3) )

		
		return
		
	def reset_translationAxesArray(self):
		"""
		Reset the translational axes value of every atom per frame while maintaining the shape of the array.
		""" 
		self.translationAxesArray = nmp.ndarray ( (len(self.trajSnapshots), self.universe.atoms.n_atoms, 4, 3) )
		return
	
	def reset_rotationAxesArray(self):
		"""
		Reset the rotational axes value of every atom per frame while maintaining the shape of the array.
		"""
		self.rotationAxesArray = nmp.ndarray ( (len(self.trajSnapshots), self.universe.atoms.n_atoms, 4, 3) )
		return

	def update_translationAxesArray_at(self, arg_frame, arg_atomList, arg_pAxes, arg_orig):
		"""
		Update the translational axes at a given frame for the selected atoms with the input values.
		"""
		self.translationAxesArray[arg_frame,arg_atomList] = nmp.vstack((arg_pAxes, arg_orig))	   
		return
	
	def update_rotationAxesArray_at(self, arg_frame, arg_atomList, arg_pAxes, arg_orig):
		"""
		Update the rotational axes at a given frame for the selected atoms with the input values.
		"""
		self.rotationAxesArray[arg_frame,arg_atomList] = nmp.vstack((arg_pAxes, arg_orig))   
		return
	

	def update_localCoords(self, arg_type, arg_atomList):
		if arg_type.upper() not in ["T","R"]:
			raise ValueError('Type {} for axes system does not exist. Chose T(translation) or R(rotation)'.format(arg_type))
			return
		
		if arg_type.upper() == "T":
			for iFrame in range(len(self.trajSnapshots)):
				for iAtom in arg_atomList:
					#translation
					self.localCoords[iFrame,iAtom] = self._labCoords[iFrame,iAtom] - self.translationAxesArray[iFrame,iAtom][-1,]
					self.localCoords[iFrame,iAtom] = self.translationAxesArray[iFrame,iAtom][0:3,] @ self.localCoords[iFrame,iAtom]
		elif arg_type.upper() == "R":
			for iFrame in range(len(self.trajSnapshots)):
				for iAtom in arg_atomList:
					#rotation
					self.localCoords[iFrame,iAtom] = self._labCoords[iFrame,iAtom] - self.rotationAxesArray[iFrame,iAtom][-1,]
					self.localCoords[iFrame,iAtom] = self.rotationAxesArray[iFrame,iAtom][0:3,] @ self.localCoords[iFrame,iAtom]

		return
	#END

	def update_localCoords_of_all_atoms(self, arg_type):
		allAtoms = nmp.arange(self.universe.atoms.n_atoms)
		self.update_localCoords(arg_type, allAtoms)
		return
	#END

	def update_localForces(self, arg_type, arg_atomList):
		if arg_type.upper() not in ["T","R"]:
			raise ValueError('Type {} for axes system does not exist. Choose T(translation) or R(rotation)'.format(arg_type))
			return
	
		if arg_type.upper() == "T":
			for iFrame in range(len(self.trajSnapshots)):
				for iAtom in arg_atomList:
					self.localForces[iFrame, iAtom] = self.translationAxesArray[iFrame, iAtom][0:3,] @ self._labForces[iFrame, iAtom]
		elif arg_type.upper() == "R":
			for iFrame in range(len(self.trajSnapshots)):
				for iAtom in arg_atomList:
					self.localForces[iFrame, iAtom] = self.rotationAxesArray[iFrame, iAtom][0:3,] @ self._labForces[iFrame, iAtom]
								
		return
	#END

	def update_localForces_of_all_atoms(self, arg_type):
		allAtoms = nmp.arange(self.universe.atoms.n_atoms)
		self.update_localForces(arg_type, allAtoms)
		return
	#END

	def  get_minmax(self, arg_token, arg_indices, arg_frame = -1, arg_local=False):
		"""
		Return the minimum and maximum value of an array with name given  by 
		`arg_token` at frame `arg_frame` for selected indices.
		"""
		if arg_frame == -1:
			# assign the last frame
			arg_frame = self.frameIndices[-1]

		arg_token = arg_token.upper()
		if arg_token in ["X", "FX"]:
			axis = 0
		elif arg_token in ["Y", "FY"]:
			axis = 1
		elif arg_token in ["Z",  "FZ"]:
			axis = 2
		else:
			raise ValueError(f"Bad array name provided for  minmax operation. ({arg_token})")

		arrType = 0
		if arg_token in ['X', 'Y', 'Z']:
			arrType = 1
			if arg_local:
				arrType = 2
		elif arg_token in ['FX', 'FY', 'FZ']:
			arrType = 3
			if arg_local:
				arrType = 4

		try:
			assert(arrType > 0)
		except:
			raise AssertionError(f'Bad array type assigned for minmax operation.({arrType})')

		if arrType == 1:
			retArr = [v3[axis] for v3 in self._labCoords[arg_frame][arg_indices]]
		elif arrType == 2:
			retArr = [v3[axis] for v3 in self.localCoords[arg_frame][arg_indices]]
		elif arrType == 3:
			retArr = [v3[axis] for v3 in self._labForces[arg_frame][arg_indices]]
		elif arrType == 4:
			retArr = [v3[axis] for v3 in self.localForces[arg_frame][arg_indices]]

		return (nmp.min(retArr), nmp.max(retArr))
#END

	def get_center_of_mass(self, arg_atomList, arg_frame):
		""" compute and return the center of mass of the input atoms in the lab frame"""
		com = nmp.zeros((3))

		totalMass = nmp.sum(nmp.asarray([self.universe.atoms.masses[iAtom] for iAtom in arg_atomList]))

		for iAtom in arg_atomList:
			iMassCoord = self.universe.atoms.masses[iAtom] * self._labCoords[arg_frame][iAtom]
			com = com + iMassCoord

		com /= totalMass
		return com

		# replace this with 
		# u.select_atoms("all").center_of_mass()
	#END


	def get_moment_of_inertia_tensor_lab(self, arg_atomList, arg_frame):
		""" return the 3 x 3 moment of inertia tensor for the body defined by the 
		collection of atoms in the input atom List. The MOIs are computed in 
		the lab frame after subtracting the COM of the atom selection from their lab coordinates.

		Moment of inertia tensor expression taken from MDanalysis website. It agrees with
		the algebric expression provided elsewhere"""

		# first compute the local coords for each atom in the input list in that frame
		# store them
		atomLocalCoords = nmp.ndarray( (len(arg_atomList),3) )
		atomMasses = nmp.zeros( (len(arg_atomList)) )

		atomsCOM = self.get_center_of_mass(arg_atomList, arg_frame)

		atIdx = 0
		for iAtom in arg_atomList:
			iLabCoord = self._labCoords[arg_frame][iAtom]

			# deduct com
			atomLocalCoords[atIdx] = iLabCoord - atomsCOM

			atomMasses[atIdx] = self.universe.atoms.masses[iAtom]
			atIdx += 1


		# generate the  MOI tensor
		tensor = nmp.zeros((3,3))

		# using standard cartesian basis for the lab frame
		basis = nmp.identity(3)

		sumDyadicCoordBasis = nmp.zeros((3,3))
		for iAxis in range(3):
			sumDyadicCoordBasis = nmp.add(sumDyadicCoordBasis, nmp.outer(basis[iAxis,], basis[iAxis,]))

		for atIdx in range(len(arg_atomList)):
			iMass = atomMasses[atIdx]
			iLocalCoord = atomLocalCoords[atIdx]

			iDyadicLocalCoord = nmp.outer(iLocalCoord, iLocalCoord)
			
			iMatrix = nmp.dot(iLocalCoord,iLocalCoord) * sumDyadicCoordBasis
			iMatrix = nmp.subtract(iMatrix,iDyadicLocalCoord)
			iMatrix = iMass * iMatrix
			
			tensor = nmp.add(tensor, iMatrix)


		return tensor
	#END    

	def get_principal_axes(self, arg_atomList, arg_frame, arg_sorted):
		""" 
		Returns a pricipal moments of inertia and a 3 x 3 matrix 
		with each row as the principal axes sorted in the descending order of the 
		corresponding eigen values (principal moments of inertia).
		"""

		#compute the moment of inertia for the imput list
		momentOfInertiaTensor = self.get_moment_of_inertia_tensor_lab(arg_atomList = arg_atomList, arg_frame = arg_frame)

		#diagonlaize
		try:
			pMomentsOfInertia, pAxes = Utils.diagonalize(momentOfInertiaTensor)
		except ComplexWarning:
			print("Moments of Inertia : ", pMomentsOfInertia)
			print("Principal Axes matrix (cols are eigen vectors):")
			print(pAxes)
			
		# IMPORTANT :::: Columns of pAxes are the eigen vectors and therefore the pricipal axes
		# When returning, return the transpose so that the rows are the pricipal axes
		# and the format is consistent with 4x3 matrix convention used for coordinate frames
		# thoughout.
		if arg_sorted:
			#sorted order (descending)
			sortOrder = nmp.argsort(-pMomentsOfInertia)

			#create a new array of the sme shape as pAxes and copy columns from it in the order in sortOrder
			pAxesSorted = nmp.zeros(nmp.shape(pAxes))

			# transposing here
			for i,j in zip(range(3),sortOrder):
				pAxesSorted[i,:] = pAxes[:,j]

			return -nmp.sort(-pMomentsOfInertia), pAxesSorted

		else:
			return pMomentsOfInertia, nmp.transpose(pAxes)

	def add_dihedral(self, arg_dihedral):
		# add the input dihedral to the value list associated with the indices of the atoms that form it.
		for atIdx in arg_dihedral.atomList:
			# Utils.printflush('Adding dihedral {} to the list for atom {}'.format(arg_dihedral, atIdx))
			self.dihedralTable[atIdx].add(arg_dihedral)
		
		self.dihedralArray.add(arg_dihedral)


import numpy as nmp
import sys

from CodeEntropy.FunctionCollection import CustomFunctions as CF

# output Forces
def write_atomic_forces( arg_baseMolecule, arg_hostDataContainer, arg_outFile ):
	""" function that prints the 
	1. average
	2. SD
	of the forces felt by each atom
	"""

	numFrames = len(arg_hostDataContainer.trajSnapshots)
	for iAtom in arg_baseMolecule.atomIndexArray:
		if not arg_baseMolecule.isHydrogenArray[iAtom]:
			iAtomname = arg_baseMolecule.atomNameArray[iAtom]
			iResid  = arg_baseMolecule.atomResidueIdxArray[iAtom]
			iResname = arg_baseMolecule.resnameArray[iResid]
			iChain = "X"

			iForces = arg_hostDataContainer._labForces[:,iAtom]
			try:
				assert(nmp.shape(iForces) == (numFrames, 3))
			except:
				CF.printflush('Requested shape ', nmp.shape(iForces), 'does not allow proper broadcasting of the array of forces from the main data containter.')

			#coordinates
			iAvgX = nmp.mean(nmp.asarray([arg_hostDataContainer._labCoords[:, iAtom, 0]]))
			iAvgY = nmp.mean(nmp.asarray([arg_hostDataContainer._labCoords[:, iAtom, 1]]))
			iAvgZ = nmp.mean(nmp.asarray([arg_hostDataContainer._labCoords[:, iAtom, 2]]))
			
			# forces
			iForceMagnitudes = nmp.asarray([ nmp.linalg.norm(iForces[i]) for i in range(numFrames)])

			# average force
			iAvgForce = nmp.mean(iForceMagnitudes)

			# std dev of forces
			iSDevForce = nmp.std(iForceMagnitudes)

			#pdb format string
			atomStr = '{:<6s}{:>5d} {:>4s} {:3s} {:1s}{:>4d}    {:>8.3f}{:>8.3f}{:>8.3f}{:>6.2f}{:>6.2f}'.format("ATOM", iAtom, iAtomname, iResname, iChain, iResid, iAvgX, iAvgY, iAvgZ, iAvgForce, iSDevForce )
			CF.printOut(arg_outFile, atomStr)

	return
#END



import numpy as nmp
from CodeEntropy.FunctionCollection import CustomFunctions as CF 

def transform_coordinates(arg_origVector, arg_oldBase, arg_newBase):
	"""Given a 3D vector (origVector; V) with components in the old_base (3,3 matrix), 
	transform it to a new 3D vector (return value) with components in the new_base (3x3 matrix).
	A 3X3 transformation matrix TM is created which gives new vector components V' = TM @ V. 
	TM[i,j] = dot(axis{j_new}, axis{i_old}).

	NOTE: Axes in both coordinate system have to be orthogonal"""

	#transformation matrix
	TM = [[nmp.dot(j,i) for i in arg_newBase] for j in arg_oldBase]

	return TM @ arg_origVector

#END


def get_cartesian_midpoint3D(arg_point1, arg_point2):
	""" Get the cartesian midpoint of the line joining two 3D points """
	return nmp.multiply(0.5, nmp.add(arg_point1, arg_point2))

#END

def generate_orthonormal_axes_system(arg_coord1, arg_coord2, arg_coord3):
	"""Generate a 3 x 3 matrix with columns as axes of the returning coordinate system. 
	They must be orthonormal (assert).
	Axis 1: Along the C-N line, from altitude bisector to C
	Axis 2: from altitde bisecting point of C-N to Ca
	Axis 3: Cross of Axis1 and Axis 2

	Return the basis and the origin"""

	basis = nmp.ndarray((3,3))

	r1 = arg_coord1
	r2 = arg_coord2
	r3 = arg_coord3

	r31 = r3 - r1
	r21 = r2 - r1

	#derived formula
	rAltitude = nmp.dot(r31, r21) * r21
	rAltitude /= nmp.linalg.norm(r21)**2
	rAltitude += r1

	# Axis 0
	axis0 = r2 - rAltitude
	axis0 /= nmp.linalg.norm(axis0)
	basis[0,:] = axis0


	# Axis 1
	axis1 = r3 - rAltitude
	axis1 /= nmp.linalg.norm(axis1)
	basis[1,:] = axis1

	# Axis 2
	# axis2 = nmp.cross(axis0, axis1)
	axis2 = CF.cross_product(axis0, axis1)
	axis2 /= nmp.linalg.norm(axis2)
	basis[2,:] = axis2

	# due to FP limitations, the check will verly likely fail
	# assert(nmp.dot(axis1, axis2) == 0)

	return basis, rAltitude

#END

def get_altitude(arg_lineEndPoint1, arg_lineEndPoint2, arg_sourcePoint):
	"""
	For the three points in the input, 
	2 form a line segement and 
	the thrid one is the point from
	where the altitude on the line segement is computed.

	Returns te altitude vector"""


	r1 = arg_lineEndPoint1
	r2 = arg_lineEndPoint2
	r3 = arg_sourcePoint

	r31 = r3 - r1
	r21 = r2 - r1

	#derived formula
	rAltitude = nmp.dot(r31, r21) * r21
	rAltitude /= nmp.linalg.norm(r21)**2
	rAltitude = r31 - rAltitude

	return rAltitude

# END


def get_sphCoord_axes(arg_r):
	""" For a given vector in space, treat it is a radial vector rooted at 0,0,0 and 
	derive a curvilinear coordinate system according to the rules of polar spherical 
	coordinates"""

	x2y2 = arg_r[0]**2 + arg_r[1]**2
	r2 = x2y2 + arg_r[2]**2

	if x2y2 != 0.:
		sinTheta = nmp.sqrt(x2y2/r2)	
		cosTheta = arg_r[2]/nmp.sqrt(r2)

		sinPhi = arg_r[1]/nmp.sqrt(x2y2)
		cosPhi = arg_r[0]/nmp.sqrt(x2y2)
		
	else:
		sinTheta = 0.
		cosTheta = 1

		sinPhi = 0.
		cosPhi = 1

	# if abs(sinTheta) > 1 or abs(sinPhi) > 1:
	#	 print('Bad sine : T {} , P {}'.format(sinTheta, sinPhi))

	# cosTheta = nmp.sqrt(1 - sinTheta*sinTheta)
	# cosPhi = nmp.sqrt(1 - sinPhi*sinPhi)
	
	# print('{} {} {}'.format(*arg_r))
	# print('Sin T : {}, cos T : {}'.format(sinTheta, cosTheta))
	# print('Sin P : {}, cos P : {}'.format(sinPhi, cosPhi))

	sphericalBasis = nmp.zeros((3,3))
	
	# r^
	sphericalBasis[0,:] = nmp.asarray([sinTheta*cosPhi, sinTheta*sinPhi, cosTheta])
	
	# Theta^
	sphericalBasis[1,:] = nmp.asarray([cosTheta*cosPhi, cosTheta*sinPhi, -sinTheta])
	
	# Phi^
	sphericalBasis[2,:] = nmp.asarray([-sinPhi, cosPhi, 0.])
	
	return sphericalBasis
	
# END

def compute_dihedral(arg_ri, arg_rj, arg_rk, arg_rl):
	""" 
	Return the dihedral angle in degrees formed by the atoms with the input coordinates.
	The formula to compute was obtained from: 
	https://www.math.fsu.edu/~quine/MB_10/6_torsion.pdf
	
	"""

	# define the directed line segments (vectors)
	a = nmp.subtract(arg_rj, arg_ri) # i->j
	b = nmp.subtract(arg_rk, arg_rj) # j->k
	c = nmp.subtract(arg_rl, arg_rk) # k->l

	# Complex number formalism for finding the angle
	# Z = X + jY
	X = -nmp.dot(a,c) * nmp.dot(b,b)
	X += ( nmp.dot(a,b) * nmp.dot(b,c) )

	Y = nmp.dot(a, nmp.cross(b,c))
	Y *= nmp.linalg.norm(b)

	# dihedral angle (in degrees)
	dih = nmp.angle(complex(X,Y), deg=True)

	return dih
# END



import MDAnalysis as mda
import pickle
from MDAnalysis.analysis.base import AnalysisFromFunction
from MDAnalysis.coordinates.memory import MemoryReader

def new_U_select_frame(u, start=None, end=None, step=1):
	"""Create a reduced universe by dropping frames according to user selection

	Parameters
	----------
	u : MDAnalyse.Universe
		A Universe object will all topology, dihedrals,coordinates and force information.
	start : int or None, Optional, default: None
		Frame id to start analysis. Default None will start from frame 0
	end : int or None, Optional, default: None
		Frame id to end analysis. Default None will end at last frame
	step : int, Optional, default: 1
		Steps between frame.

	Returns
	-------
		u2 : MDAnalysis.Universe
			reduced universe
	"""
	if start == None:
		start = 0
	if end == None:
		end = len(u.trajectory)
	select_atom = u.select_atoms('all')
	coordinates = AnalysisFromFunction(lambda ag: ag.positions.copy(), select_atom).run().results['timeseries'][start:end:step]
	forces = AnalysisFromFunction(lambda ag: ag.forces.copy(), select_atom).run().results['timeseries'][start:end:step]
	dimensions = AnalysisFromFunction(lambda ag: ag.dimensions.copy(), select_atom).run().results['timeseries'][start:end:step]
	u2 = mda.Merge(select_atom)
	u2.load_new(coordinates, format=MemoryReader, forces=forces, dimensions=dimensions)
	return u2

def new_U_select_atom(u, select_string='all'):
	"""Create a reduced universe by dropping atoms according to user selection

	Parameters
	----------
	u : MDAnalyse.Universe
		A Universe object will all topology, dihedrals,coordinates and force information.
	select_string : str, Optional, default: 'all'
		MDAnalysis.select_atoms selection string.

	Returns
	-------
		u2 : MDAnalysis.Universe
			reduced universe

	"""
	select_atom = u.select_atoms(select_string)
	coordinates = AnalysisFromFunction(lambda ag: ag.positions.copy(), select_atom).run().results['timeseries']
	forces = AnalysisFromFunction(lambda ag: ag.forces.copy(), select_atom).run().results['timeseries']
	dimensions = AnalysisFromFunction(lambda ag: ag.dimensions.copy(), select_atom).run().results['timeseries']
	u2 = mda.Merge(select_atom)
	u2.load_new(coordinates, format=MemoryReader, forces=forces, dimensions=dimensions)
	return u2        

def write_universe(u, name='default'):
	"""Write a universe to working directories as pickle
	
	Parameters
	----------
	u : MDAnalyse.Universe
		A Universe object will all topology, dihedrals,coordinates and force information.
	name : str, Optional. default: 'default'
		The name of file with sub file name .pkl

	Returns
	-------
		name : str
			filename of saved universe
	"""
	filename = f"{name}.pkl"
	pickle.dump(u, open(filename, "wb"))
	return name

def read_universe(path):
	"""read a universe to working directories as pickle

	Parameters
	----------
	path : str
		The path to file.

	Returns
	-------
		u : MDAnalysis.Universe
			A Universe object will all topology, dihedrals,coordinates and force information.
	"""
	u = pickle.load(open(path, "rb"))
	return u